from neupy.core.docs import SharedDocs
from neupy.core.properties import WithdrawProperty
from neupy.utils import format_data


__all__ = ('SupervisedLearningMixin', 'LazyLearningMixin')


class SupervisedLearningMixin(object):
    """
    Mixin for Supervised Neural Network algorithms.

    Methods
    -------
    train(input_train, target_train, input_test=None, target_test=None,\
    epochs=100, epsilon=None)
        Train network. You can control network's training procedure
        with ``epochs`` and ``epsilon`` parameters.
        The ``input_test`` and ``target_test`` should be presented
        both in case of you need to validate network's training
        after each iteration.
    """
    def train(self, input_train, target_train, input_test=None,
              target_test=None, epochs=100, epsilon=None,
              summary='table'):

        is_test_data_partialy_missed = (
            (input_test is None and target_test is not None) or
            (input_test is not None and target_test is None)
        )

        if is_test_data_partialy_missed:
            raise ValueError("Input and target test samples are missed. "
                             "They must be defined together or none of them.")

        input_train = format_data(input_train)
        target_train = format_data(target_train)

        if input_test is not None:
            input_test = format_data(input_test)

        if target_test is not None:
            target_test = format_data(target_test)

        return super(SupervisedLearningMixin, self).train(
            input_train=input_train, target_train=target_train,
            input_test=input_test, target_test=target_test,
            epochs=epochs, epsilon=epsilon,
            summary=summary
        )


class LazyLearningMixin(SharedDocs):
    """
    Mixin for lazy learning Neural Network algorithms.

    Notes
    -----
    - Network uses lazy learning which mean that network doesn't
      need iterative training. It just stores parameters
      and use them to make a predictions.

    Methods
    -------
    train(input_train, target_train, copy=True)
        Network just stores all the information about the data and use
        it for the prediction. Parameter ``copy`` copies input data
        before saving it inside the network.
    """
    step = WithdrawProperty()
    show_epoch = WithdrawProperty()
    shuffle_data = WithdrawProperty()
    train_end_signal = WithdrawProperty()
    epoch_end_signal = WithdrawProperty()

    def __init__(self, *args, **kwargs):
        self.input_train = None
        self.target_train = None
        super(LazyLearningMixin, self).__init__(*args, **kwargs)

    def train(self, input_train, target_train):
        self.input_train = input_train
        self.target_train = target_train

        if input_train.shape[0] != target_train.shape[0]:
            raise ValueError("Number of samples in the input and target "
                             "datasets are different")



from itertools import product
from contextlib import contextmanager

from neupy.layers.utils import preformat_layer_shape
from .utils import join, is_sequential
from .graph import LayerGraph


__all__ = ('LayerConnection', 'BaseConnection', 'ParallelConnection')


class GlobalConnectionState(dict):
    def __setitem__(self, key, value):
        return super(GlobalConnectionState, self).__setitem__(id(key), value)

    def __getitem__(self, key):
        return super(GlobalConnectionState, self).__getitem__(id(key))

    def __contains__(self, key):
        return super(GlobalConnectionState, self).__contains__(id(key))


class BaseConnection(object):
    """
    Base class from chain connections.

    Attributes
    ----------
    graph : LayerGraph
        Reference to the graph that contains all lations
        between layers specified in the connection.

    training_state : bool
        Defines state of the connection. Defaults to ``True``.

    input_layers : list of layers
        List of connection's input layers.

    output_layers : list of layers
        List of connection's output layers.
    """
    left_states = GlobalConnectionState()
    right_states = GlobalConnectionState()

    def __init__(self):
        self.training_state = True
        self.graph = LayerGraph()

        self.input_layers = [self]
        self.output_layers = [self]

    @classmethod
    def connect(self, left, right):
        """
        Make connection between two objects.
        """
        main_left, main_right = left, right

        if left in self.right_states and left not in self.left_states:
            left = self.right_states[left]

        if right in self.left_states and right not in self.right_states:
            right = self.left_states[right]

        connection = LayerConnection(left, right)

        self.left_states[main_left] = connection
        self.right_states[main_right] = connection

        return connection

    def __gt__(self, other):
        return self.__class__.connect(self, other)

    def __lt__(self, other):
        return self.__class__.connect(other, self)

    def __rshift__(self, other):
        return self.__class__.connect(self, other)

    def __lshift__(self, other):
        return self.__class__.connect(other, self)

    def __iter__(self):
        yield self

    def output(self, input_value):
        """
        Return output base on the input value.

        Parameters
        ----------
        input_value
        """
        raise NotImplementedError

    def initialize(self):
        """
        Initialize connection
        """

    @contextmanager
    def disable_training_state(self):
        """
        Disable connection's training state
        """
        self.training_state = False
        yield
        self.training_state = True


def make_common_graph(left_layer, right_layer):
    """
    Makes common graph for two layers that exists
    in different graphs.

    Parameters
    ----------
    left_layer : layer
    right_layer : layer

    Returns
    -------
    LayerGraph instance
        Graph that contains both layers and their connections.
    """
    graph = LayerGraph.merge(left_layer.graph, right_layer.graph)

    for layer in graph.forward_graph.keys():
        layer.graph = graph

    return graph


def topological_sort(graph):
    """
    Repeatedly go through all of the nodes in the graph, moving each of
    the nodes that has all its edges resolved, onto a sequence that
    forms our sorted graph. A node has all of its edges resolved and
    can be moved once all the nodes its edges point to, have been moved
    from the unsorted graph onto the sorted one.

    Parameters
    ----------
    graph : dict
        Dictionary that has graph structure.

    Raises
    ------
    RuntimeError
        If graph has cycles.

    Returns
    -------
    list
        List of nodes sorted in topological order.
    """
    sorted_nodes = []
    graph_unsorted = graph.copy()

    while graph_unsorted:
        acyclic = False

        for node, edges in list(graph_unsorted.items()):
            if all(edge not in graph_unsorted for edge in edges):
                acyclic = True
                del graph_unsorted[node]
                sorted_nodes.append(node)

        if not acyclic:
            raise RuntimeError("A cyclic dependency occurred")

    return sorted_nodes


class ParallelConnection(BaseConnection):
    """
    Connection between separate layer networks in parallel.
    Each network has it's own input and out layers.

    Parameters
    ----------
    connections : list of list or connection
        List of networks.

    Attributes
    ----------
    connections : list of connections
        List of networks.

    input_layers : list of connections
        Combined list of all input layers that each
        parallel connection has.

    output_layers : list of connections
        Combined list of all output layers that each
        parallel connection has.
    """
    def __init__(self, connections):
        from neupy.layers.base import ResidualConnection

        super(ParallelConnection, self).__init__()

        self.connections = []
        self.input_layers = []
        self.output_layers = []

        for layers in connections:
            if isinstance(layers, BaseConnection):
                connection = layers
            elif not layers:
                connection = ResidualConnection()
            else:
                connection = join(*layers)

            self.connections.append(connection)
            self.input_layers.extend(connection.input_layers)
            self.output_layers.extend(connection.output_layers)

    @property
    def input_shape(self):
        """
        Returns input shape per each network parallel
        connection.
        """
        input_shapes = []
        for connection in self.connections:
            input_shapes.append(connection.input_shape)
        return input_shapes

    @property
    def output_shape(self):
        """
        Returns output shape per each network parallel
        connection.
        """
        output_shapes = []
        for connection in self.connections:
            output_shapes.append(connection.output_shape)
        return output_shapes

    def initialize(self):
        """
        Initialize all connections.
        """
        for connection in self.connections:
            connection.initialize()

    def output(self, *input_values):
        """
        Compute outputs per each network in parallel
        connection.

        Parameters
        ----------
        *input_values

        Returns
        -------
        list
        """
        n_inputs = len(input_values)
        n_connections = len(self.connections)

        if n_inputs == 1:
            input_values = list(input_values) * n_connections

        elif n_inputs != n_connections:
            raise ValueError("Expected {} input values for parallel "
                             "connection, got {}".format(n_connections,
                                                         n_inputs))

        outputs = []
        for input_value, connection in zip(input_values, self.connections):
            connection_output = connection.output(input_values)
            outputs.append(connection_output)

        return outputs

    @contextmanager
    def disable_training_state(self):
        """
        Disable training state for all layers in all
        connections.
        """
        for connection in self.connections:
            for layer in connection:
                layer.training_state = False

        yield

        for connection in self.connections:
            for layer in connection:
                layer.training_state = True

    def __iter__(self):
        for connection in self.connections:
            yield connection


class LayerConnection(BaseConnection):
    """
    Make connection between layers.

    Parameters
    ----------
    left : layer, connection or list of connections
    right : layer, connection or list of connections
    """
    def __init__(self, left, right):
        super(LayerConnection, self).__init__()

        if isinstance(left, (list, tuple)):
            left = ParallelConnection(left)

        if isinstance(right, (list, tuple)):
            right = ParallelConnection(right)

        self.left = left
        self.right = right

        layers = product(left.output_layers, right.input_layers)
        for left_output, right_input in layers:
            self.full_graph = make_common_graph(left_output, right_input)

        self.full_graph.connect_layers(left.output_layers, right.input_layers)

        self.input_layers = self.left.input_layers
        self.output_layers = self.right.output_layers

        # Generates subgraph that contains only connections
        # between specified input and output layers
        self.graph = self.full_graph.subgraph(self.input_layers,
                                              self.output_layers)

    @property
    def input_shape(self):
        """
        Connection's input shape/shapes.

        Returns
        -------
        list of tuples, tuple or None
            - List of tuples: in case if there are more than
              one input layer. Each tuple is a shape of the
              specific input layer.

            - tuple: in case if there is one input layer.
              Tuple object defines input layer's shape.
        """
        # Cannot save them during initialization step,
        # because input shape can be modified later
        if len(self.input_layers) == 1:
            input_layer = self.input_layers[0]
            return input_layer.input_shape

        input_shapes = []
        for input_layer in self.input_layers:
            input_shapes.append(input_layer.input_shape)

        return input_shapes

    @property
    def output_shape(self):
        """
        Connection's output shape/shapes.

        Returns
        -------
        list of tuples, tuple or None
            - List of tuples: in case if there are more than
              one input layer. Each tuple is a shape of the
              specific input layer.

            - tuple: in case if there is one input layer.
              Tuple object defines input layer's shape.
        """
        # Cannot save them during initialization step,
        # because input shape can be modified later
        if len(self.output_layers) == 1:
            output_layer = self.output_layers[0]
            return output_layer.output_shape

        output_shapes = []
        for output_layer in self.output_layers:
            output_shapes.append(output_layer.output_shape)

        return output_shapes

    def initialize(self):
        """
        Initialize all layers in the connection.
        """
        for layer in self:
            layer.initialize()

    def output(self, *input_values):
        """
        Propagate input values through all layers in the
        connections and returns output from the final layers.

        Parameters
        ----------
        *input_values
            - Input values can be Theano variables or
              array-like objects

            - Dictionary inputs should have key that
              define input layer and value is a variables
              that needs to be propagated through all layers.

        Returns
        -------
        Theano expression
        """
        subgraph = self.graph
        n_inputs = len(input_values)

        if n_inputs == 1 and not isinstance(input_values[0], dict):
            input_value = input_values[0]
            new_input_values = {}

            for input_layer in self.input_layers:
                new_input_values[input_layer] = input_value

            input_values = [new_input_values]

        return subgraph.propagate_forward(*input_values)

    @contextmanager
    def disable_training_state(self):
        """
        Disable training state for all layers in the
        connection.
        """
        for layer in self:
            layer.training_state = False

        yield

        for layer in self:
            layer.training_state = True

    def __len__(self):
        return len(self.graph.forward_graph)

    def __iter__(self):
        backward_graph = self.graph.backward_graph
        for layer in topological_sort(backward_graph):
            yield layer

    def __repr__(self):
        n_layers = len(self)

        if n_layers > 5 or not is_sequential(self):
            conn = '{} -> [... {} layers ...] -> {}'.format(
                preformat_layer_shape(self.input_shape),
                n_layers,
                preformat_layer_shape(self.output_shape)
            )
        else:
            conn = ' > '.join([repr(layer) for layer in self])

        return conn



import abc
import time
import types

import six
import theano
import theano.sparse
import theano.tensor as T

from neupy import layers
from neupy.utils import AttributeKeyDict, asfloat, format_data
from neupy.layers.utils import preformat_layer_shape
from neupy.layers.connections import LayerConnection, is_sequential
from neupy.exceptions import InvalidConnection
from neupy.helpers import table
from neupy.core.properties import ChoiceProperty
from neupy.algorithms.base import BaseNetwork
from neupy.algorithms.utils import parameter_values
from neupy.algorithms.learning import SupervisedLearningMixin
from .gd import errors


__all__ = ('ConstructibleNetwork',)


def does_layer_accept_1d_feature(layer):
    """
    Check if 1D feature values are valid for the layer.

    Parameters
    ----------
    layer : object

    Returns
    -------
    bool
    """
    return (layer.output_shape == (1,))


def generate_layers(layers_sizes):
    """
    Create from list of layer sizes basic linear network.

    Parameters
    ----------
    layers_sizes : list or tuple
        Ordered list of network connection structure.

    Returns
    -------
    LayerConnection
        Constructed connection.
    """
    n_layers = len(layers_sizes)

    if n_layers <= 1:
        raise ValueError("Cannot generate network that "
                         "has less than two layers")

    input_layer_size = layers_sizes.pop(0)
    connection = layers.Input(input_layer_size)

    for output_size in layers_sizes:
        next_layer = layers.Sigmoid(output_size)
        connection = LayerConnection(connection, next_layer)

    return connection


def clean_layers(connection):
    """
    Clean layers connections and format transform them into one format.
    Also this function validate layers connections.

    Parameters
    ----------
    connection : list, tuple or object
        Layers connetion in different formats.

    Returns
    -------
    object
        Cleaned layers connection.
    """

    if isinstance(connection, tuple):
        connection = list(connection)

    if all(isinstance(element, int) for element in connection):
        connection = generate_layers(connection)

    islist = isinstance(connection, list)
    layer_types = (layers.BaseLayer, LayerConnection)

    if islist and isinstance(connection[0], layer_types):
        connection = layers.join(*connection)

    return connection


def create_input_variable(input_layer, name):
    """
    Create input variable based on input layer information.

    Parameters
    ----------
    input_layer : object
    name : str

    Returns
    -------
    Theano variable
    """
    dim_to_variable_type = {
        2: T.matrix,
        3: T.tensor3,
        4: T.tensor4,
    }

    # Shape doesn't include batch size dimension,
    # that's why we need to add one
    ndim = len(input_layer.input_shape) + 1

    if ndim not in dim_to_variable_type:
        raise ValueError("Layer's input needs to be 2, 3 or 4 dimensional. "
                         "Found {} dimensions".format(ndim))

    variable_type = dim_to_variable_type[ndim]
    return variable_type(name)


def create_output_variable(error_function, name):
    """
    Create output variable based on error function.

    Parameters
    ----------
    error_function : function
    name : str

    Returns
    -------
    Theano variable
    """
    # TODO: Solution is not user friendly. I need to find
    # better solution later.
    if hasattr(error_function, 'expected_dtype'):
        network_output_dtype = error_function.expected_dtype
    else:
        network_output_dtype = T.matrix

    return network_output_dtype(name)


class ErrorFunctionProperty(ChoiceProperty):
    """
    Property that helps select error function from
    available or define a new one.

    Parameters
    ----------
    {ChoiceProperty.Parameters}
    """
    def __set__(self, instance, value):
        if isinstance(value, types.FunctionType):
            return super(ChoiceProperty, self).__set__(instance, value)
        return super(ErrorFunctionProperty, self).__set__(instance, value)

    def __get__(self, instance, value):
        founded_value = super(ChoiceProperty, self).__get__(instance, value)
        if isinstance(founded_value, types.FunctionType):
            return founded_value
        return super(ErrorFunctionProperty, self).__get__(instance,
                                                          founded_value)


class BaseAlgorithm(six.with_metaclass(abc.ABCMeta)):
    """
    Base class for algorithms implemeted in Theano.

    Attributes
    ----------
    variables : dict
        Theano variables.

    methods : dict
        Compiled Theano functions.
    """
    def __init__(self, *args, **kwargs):
        super(BaseAlgorithm, self).__init__(*args, **kwargs)

        self.logs.message("THEANO", "Initializing Theano variables and "
                                    "functions.")
        start_init_time = time.time()

        self.variables = AttributeKeyDict()
        self.methods = AttributeKeyDict()

        self.init_input_output_variables()
        self.init_variables()
        self.init_methods()

        finish_init_time = time.time()
        self.logs.message("THEANO", "Initialization finished successfully. "
                          "It took {:.2f} seconds"
                          "".format(finish_init_time - start_init_time))

    @abc.abstractmethod
    def init_input_output_variables(self):
        """
        Initialize input and output Theano variables.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def init_variables(self):
        """
        Initialize Theano variables.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def init_methods(self):
        """
        Initialize Theano functions.
        """
        raise NotImplementedError


class ConstructibleNetwork(SupervisedLearningMixin, BaseAlgorithm,
                           BaseNetwork):
    """
    Class contains functionality that helps work with network that have
    constructible layers architecture.

    Parameters
    ----------
    connection : list, tuple or LayerConnection instance
        Network's architecture. There are a few ways
        to define it.

        - List of layers.
          For instance, ``[Input(2), Tanh(4), Relu(1)]``.

        - Construct layer connections.
          For instance, ``Input(2) > Tanh(4) > Relu(1)``.

        - Tuple of integers. Each integer defines Sigmoid layer
          and it's input size.  For instance,  value ``(2, 4, 1)``
          means that network has 3 layers with 2 input units,
          4 hidden units and 1 output unit.

    error : str or function
        Error/loss function. Defaults to ``mse``.

        - ``mae`` - Mean Absolute Error.

        - ``mse`` - Mean Squared Error.

        - ``rmse`` - Root Mean Squared Error.

        - ``msle`` - Mean Squared Logarithmic Error.

        - ``rmsle`` - Root Mean Squared Logarithmic Error.

        - ``categorical_crossentropy`` - Categorical cross entropy.

        - ``binary_crossentropy`` - Binary cross entropy.

        - ``binary_hinge`` - Binary hinge entropy.

        - ``categorical_hinge`` - Categorical hinge entropy.

        - Custom function which accepts two mandatory arguments.
          The first one is expected value and the second one is
          predicted value. Example:

        .. code-block:: python

            def custom_func(expected, predicted):
                return expected - predicted

    {BaseNetwork.Parameters}

    Attributes
    ----------
    {BaseNetwork.Attributes}

    Methods
    -------
    {BaseSkeleton.predict}

    {SupervisedLearningMixin.train}

    {BaseSkeleton.fit}
    """
    error = ErrorFunctionProperty(default='mse', choices={
        'mae': errors.mae,
        'mse': errors.mse,
        'rmse': errors.rmse,
        'msle': errors.msle,
        'rmsle': errors.rmsle,

        'binary_crossentropy': errors.binary_crossentropy,
        'categorical_crossentropy': errors.categorical_crossentropy,

        'binary_hinge': errors.binary_hinge,
        'categorical_hinge': errors.categorical_hinge,
    })

    def __init__(self, connection, *args, **kwargs):
        self.connection = clean_layers(connection)

        self.layers = list(self.connection)
        graph = self.connection.graph

        if len(self.connection.input_layers) != 1:
            n_inputs = len(graph.input_layers)
            raise InvalidConnection("Connection should have one input "
                                    "layer, got {}".format(n_inputs))

        if len(self.connection.output_layers) != 1:
            n_outputs = len(graph.output_layers)
            raise InvalidConnection("Connection should have one output "
                                    "layer, got {}".format(n_outputs))

        self.input_layer = graph.input_layers[0]
        self.output_layer = graph.output_layers[0]

        super(ConstructibleNetwork, self).__init__(*args, **kwargs)

    def init_input_output_variables(self):
        self.variables.update(
            network_input=create_input_variable(
                self.input_layer,
                name='algo:network/var:network-input',
            ),
            network_output=create_output_variable(
                self.error,
                name='algo:network/var:network-output',
            ),
        )

    def init_variables(self):
        network_input = self.variables.network_input
        network_output = self.variables.network_output

        train_prediction = self.connection.output(network_input)
        with self.connection.disable_training_state():
            prediction = self.connection.output(network_input)

        self.variables.update(
            step=theano.shared(
                name='algo:network/scalar:step',
                value=asfloat(self.step)
            ),
            epoch=theano.shared(
                name='algo:network/scalar:epoch',
                value=asfloat(self.last_epoch)
            ),

            prediction_func=prediction,
            train_prediction_func=train_prediction,

            error_func=self.error(network_output, train_prediction),
            validation_error_func=self.error(network_output, prediction),
        )

    def init_methods(self):
        network_input = self.variables.network_input
        network_output = self.variables.network_output

        self.methods.update(
            predict=theano.function(
                inputs=[network_input],
                outputs=self.variables.prediction_func,
                name='algo:network/func:predict'
            ),
            train_epoch=theano.function(
                inputs=[network_input, network_output],
                outputs=self.variables.error_func,
                updates=self.init_train_updates(),
                name='algo:network/func:train-epoch'
            ),
            prediction_error=theano.function(
                inputs=[network_input, network_output],
                outputs=self.variables.validation_error_func,
                name='algo:network/func:prediction-error'
            )
        )

    def init_train_updates(self):
        """
        Initialize train function update in Theano format that
        would be trigger after each training epoch.
        """
        updates = []
        for layer in self.layers:
            updates.extend(self.init_layer_updates(layer))
        return updates

    def init_layer_updates(self, layer):
        """
        Initialize train function update in Theano format that
        would be trigger after each training epoch for each layer.

        Parameters
        ----------
        layer : object
            Any layer that inherit from layers.BaseLayer class.

        Returns
        -------
        list
            Update that excaptable by ``theano.function``. There should be
            a lits that contains tuples with 2 elements. First one should
            be parameter that would be updated after epoch and the second one
            should update rules for this parameter. For example parameter
            could be a layer's weight and bias.
        """
        updates = []
        for parameter in parameter_values(layer):
            updates.extend(self.init_param_updates(layer, parameter))
        updates.extend(layer.updates)
        return updates

    def init_param_updates(self, layer, parameter):
        """
        Initialize parameter updates.

        Parameters
        ----------
        layer : object
            Any layer that inherit from BaseLayer class.
        parameter : object
            Usualy it is a weight or bias.

        Returns
        -------
        list
            List of updates related to the specified parameter.
        """
        return []

    def format_input_data(self, input_data):
        """
        Input data format is depend on the input layer
        structure.

        Parameters
        ----------
        input_data : array-like or None

        Returns
        -------
        array-like or None
            Function returns formatted array.
        """
        if input_data is not None:
            is_feature1d = does_layer_accept_1d_feature(self.input_layer)
            return format_data(input_data, is_feature1d)

    def format_target_data(self, target_data):
        """
        Target data format is depend on the output layer
        structure.

        Parameters
        ----------
        target_data : array-like or None

        Returns
        -------
        array-like or None
            Function returns formatted array.
        """
        if target_data is not None:
            is_feature1d = does_layer_accept_1d_feature(self.output_layer)
            return format_data(target_data, is_feature1d)

    def prediction_error(self, input_data, target_data):
        """
        Calculate prediction accuracy for input data.

        Parameters
        ----------
        input_data : array-like
        target_data : array-like

        Returns
        -------
        float
            Prediction error.
        """
        return self.methods.prediction_error(
            self.format_input_data(input_data),
            self.format_target_data(target_data)
        )

    def predict(self, input_data):
        """
        Return prediction results for the input data.

        Parameters
        ----------
        input_data : array-like

        Returns
        -------
        array-like
        """
        input_data = self.format_input_data(input_data)
        return self.methods.predict(input_data)

    def on_epoch_start_update(self, epoch):
        """
        Function would be trigger before run all training procedure
        related to the current epoch.

        Parameters
        ----------
        epoch : int
            Current epoch number.
        """
        super(ConstructibleNetwork, self).on_epoch_start_update(epoch)
        self.variables.epoch.set_value(epoch)

    def train(self, input_train, target_train, input_test=None,
              target_test=None, *args, **kwargs):
        """
        Trains neural network.
        """
        return super(ConstructibleNetwork, self).train(
            self.format_input_data(input_train),
            self.format_target_data(target_train),
            self.format_input_data(input_test),
            self.format_target_data(target_test),
            *args, **kwargs
        )

    def train_epoch(self, input_train, target_train):
        """
        Trains neural network over one epoch.

        Parameters
        ----------
        input_data : array-like
        target_data : array-like

        Returns
        -------
        float
            Prediction error.
        """
        return self.methods.train_epoch(input_train, target_train)

    def architecture(self):
        """
        Shows network's architecture in the terminal if
        ``verbose`` parameter is equal to ``True``.
        """
        if not is_sequential(self.connection):
            raise TypeError("You can check architecture only for sequential "
                            "connections. For other types of connections it's "
                            "better to use the `neupy.plots.layer_structure` "
                            "function.")

        self.logs.title("Network's architecture")

        values = []
        for index, layer in enumerate(self.layers, start=1):
            input_shape = preformat_layer_shape(layer.input_shape)
            output_shape = preformat_layer_shape(layer.output_shape)
            classname = layer.__class__.__name__

            values.append((index, input_shape, classname, output_shape))

        table.TableBuilder.show_full_table(
            columns=[
                table.Column(name="#"),
                table.Column(name="Input shape"),
                table.Column(name="Layer Type"),
                table.Column(name="Output shape"),
            ],
            values=values,
            stdout=self.logs.write,
        )
        self.logs.newline()

    def __repr__(self):
        return "{}({}, {})".format(self.class_name(), self.connection,
                                   self.repr_options())



__all__ = ('preformat_layer_shape', 'dimshuffle', 'iter_parameters',
           'count_parameters')


def preformat_layer_shape(shape):
    """
    Format layer's input or output shape.

    Parameters
    ----------
    shape : int or tuple

    Returns
    -------
    int or tuple
    """
    if isinstance(shape, tuple) and len(shape) == 1:
        return shape[0]
    return shape


def dimshuffle(value, ndim, axes):
    """
    Shuffle dimension based on the specified number of
    dimensions and axes.

    Parameters
    ----------
    value : Theano variable
    ndim : int
    axes : tuple, list

    Returns
    -------
    Theano variable
    """
    pattern = ['x'] * ndim

    for i, axis in enumerate(axes):
        pattern[axis] = i

    return value.dimshuffle(pattern)


def iter_parameters(layers, only_trainable=True):
    """
    Iterate through layer parameters.

    Parameters
    ----------
    layers : list of layers or connection

    Yields
    ------
    tuple
        Tuple with three ariables: (layer, attribute_name, parameter)
    """
    for layer in layers:
        for attrname, parameter in layer.parameters.items():
            if parameter.trainable or not only_trainable:
                yield layer, attrname, parameter


def count_parameters(connection):
    """
    Count number of parameters in Neural Network.

    Parameters
    ----------
    connection : list of laters or connection

    Returns
    -------
    int
        Number of parameters.
    """
    n_parameters = 0

    for _, _, parameter in iter_parameters(connection):
        parameter = parameter.get_value()
        n_parameters += parameter.size

    return n_parameters
